KEY_ATTRS = (
    'acl',
    'cache_control',
    'canned_acl',
    'content_disposition',
    'content_encoding',
    'content_type',
    'content_language',
    'encrypted',
    'expires',
    'metadata',
    'reduced_redundancy',
)

KEY_HEADERS = (
    'cache_control',
    'content_disposition',
    'content_encoding',
    'content_type',
    'content_language',
    'expires',
)

BUCKET_ATTRS = (
    'acl',
    'canned_acl',
    'cors',
    'lifecycle',
    'logging',
    'notification',
    'policy',
    'region',
    'tagging',
    'versioning',
    'website',
    'redirects',
)